import logging
from typing import Optional
from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field
from cg_lims.models.arnold.base_step import BaseStep
from cg_lims.objects import BaseAnalyte

LOG = logging.getLogger(__name__)


class ProcessUDFs(BaseModel):
    """Normalization of RNA samples for sequencing v1"""

    dilution_lot_nr: Optional[str] = Field(None, alias="Dilution buffer lot no")


class ArtifactUDFs(BaseModel):
    concentration: Optional[float] = Field(None, alias="Concentration")
    size: Optional[int] = Field(None, alias="Size (bp)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_normalization_of_samples(lims: Lims, sample_id: str, prep_id: str) -> Optional[ArnoldStep]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Normalization of RNA samples for sequencing v1",
        optional_step=True,
    )
    if not analyte.artifact:
        return None

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="normalization_for_sequencing",
        workflow="RNA"
    )



from pydantic import BaseModel


class PlacementMapHeader(BaseModel):
    process_type: str
    date: str


class PlateInfo(BaseModel):
    container_name: str
    container_type: str
    container_id: str


class WellInfo(BaseModel):
    project_name: str
    sample_name: str
    sample_id: str
    container_source: str
    container_name: str
    well_source: str
    well: str
    exta_udf_info: str
    dest_well: str
    sample_type: str



from typing import List

from genologics.lims import Lims

from .define_run_format import get_define_run_format
from .standard_make_pool_and_denature import get_standard_make_pool_and_denature
from .standard_prepare_for_sequencing import get_standard_prepare_for_sequencing
from .xp_denature_and_examp import get_xp_denature_and_examp
from .xp_load_to_flowcell import get_xp_load_to_flowcell
from .bcl_conversion_and_demultiplexing import get_bcl_conversion_and_demultiplexing
from cg_lims.models.arnold.base_step import BaseStep


def build_novaseq_step_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a Step Documents for a NovaSeq sequencing run."""

    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_define_run_format(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_standard_make_pool_and_denature(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_standard_prepare_for_sequencing(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_xp_denature_and_examp(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_xp_load_to_flowcell(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_bcl_conversion_and_demultiplexing(
            sample_id=sample_id, lims=lims, prep_id=prep_id, process_id=process_id
        ),
    ]
    return [document for document in step_documents if document is not None]



from typing import Literal
from genologics.entities import Artifact


def set_qc_fail(
    artifact: Artifact,
    value: float,
    threshold: float,
    criteria: Literal[">=", "<=", ">", "<", "==", "!="],
) -> None:
    if criteria == ">=" and value >= threshold:
        artifact.qc_flag = "FAILED"
    elif criteria == "<=" and value <= threshold:
        artifact.qc_flag = "FAILED"
    elif criteria == ">" and value > threshold:
        artifact.qc_flag = "FAILED"
    elif criteria == "<" and value < threshold:
        artifact.qc_flag = "FAILED"
    elif criteria == "==" and value == threshold:
        artifact.qc_flag = "FAILED"
    elif criteria == "!=" and value != threshold:
        artifact.qc_flag = "FAILED"



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.base_step import BaseStep


class ProcessUDFs(BaseModel):
    lot_nr_truseq_library_preparation_kit: Optional[str] = Field(
        None, alias="Lot no: TruSeq DNA PCR-Free Sample Prep Kit"
    )
    lot_nr_index: Optional[str] = Field(None, alias="Lot no: Adaptor Plate")
    lot_nr_beads: Optional[str] = Field(None, alias="Lot no: SP Beads")
    lot_nr_lucigen_library_preparation_kit: Optional[str] = Field(
        None, alias="Lot no: Lucigen prep kit"
    )
    pcr_instrument_incubation: Optional[str] = Field(None, alias="PCR machine")
    lot_nr_h2o_library_preparation: Optional[str] = Field(None, alias="Lot no: Nuclease free water")
    lot_nr_resuspension_buffer_library_preparation: Optional[str] = Field(
        None, alias="Lot no: Resuspension buffer"
    )
    methods: Optional[str] = Field(None, alias="Methods")
    atlas_version: Optional[str] = Field(None, alias="Atlas Version")
    lot_nr_etoh_library_preparation: Optional[str] = Field(None, alias="Ethanol lot")


class ArtifactUDFs(BaseModel):
    finished_library_concentration: Optional[float] = Field(None, alias="Concentration")
    finished_library_concentration_nm: Optional[float] = Field(None, alias="Concentration (nM)")
    finished_library_size: Optional[int] = Field(None, alias="Size (bp)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_end_repair(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="End repair Size selection A-tailing and Adapter ligation (TruSeq PCR-free DNA)",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="end_repair",
        workflow="WGS",
    )



"""Scripts to check the process artifact udfs are set.
"""
import logging
import sys
from typing import List
import click
from genologics.entities import Process, Artifact
from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def check_udfs(artifacts: List[Artifact], artifact_udfs: List[str]) -> None:
    """Check that process artifact udfs are set."""

    warning = []
    for udf in artifact_udfs:
        missing_udfs = [artifact.id for artifact in artifacts if artifact.udf.get(udf) is None]
        if missing_udfs:
            warning.append(f"Udf: {udf} missing for artifacts: {missing_udfs}.")
    if warning:
        LOG.warning(" ".join(warning))
        raise MissingUDFsError(message=" ".join(warning))
    LOG.info("Artifact udfs were all set.")


@click.command()
@options.artifact_udfs()
@options.measurement(help="Set to True if you want to check measurement udfs.")
@options.input(help="Set to True if you want to check the input artifact udfs.")
@click.pass_context
def check_artifact_udfs(
    ctx: click.Context, artifact_udfs: List[str], input: bool, measurement: bool
):
    """Script to check that artifact udfs are set."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process: Process = ctx.obj["process"]
    try:
        artifacts = get_artifacts(process=process, input=input, measurement=measurement)
        check_udfs(artifacts=artifacts, artifact_udfs=artifact_udfs)
        click.echo("Artifact udfs were checked.")
    except LimsError as e:
        sys.exit(e.message)



from genologics.entities import Process

from cg_lims.EPPs.arnold.flow_cell import build_flow_cell_document
from cg_lims.get.artifacts import get_output_artifacts
from cg_lims.models.arnold.flow_cell import FlowCell

from tests.conftest import server


def test_load_flowcell(lims, flow_cell_fixture):
    # GIVEN: A lims with a process: "24-314318" (NovaSeq Run).

    server("novaseq_run")

    process = Process(lims=lims, id="24-314318")
    lanes = get_output_artifacts(
        process=process, output_generation_type=["PerInput"], lims=lims, output_type="ResultFile"
    )

    # WHEN running build_flow_cell_document
    flow_cell: FlowCell = build_flow_cell_document(process=process, lanes=lanes)

    # THEN assert the flow_cell document with its lanes was created.
    flow_cell_dict = flow_cell.dict()
    lanes = flow_cell_dict.pop("lanes")
    fixture_lanes = flow_cell_fixture.pop("lanes")
    for lane in fixture_lanes:
        assert lane in lanes
    assert flow_cell_dict == flow_cell_fixture



from typing import List

from genologics.lims import Lims

from .reception_control import get_sample_artifact_fields
from .microbial_library_prep_nextera import get_library_prep_nextera
from .normalization_of_microbial_samples import get_normalization_of_mictobial_samples
from .buffer_exchange import get_buffer_exchange
from .normailzation_of_microbial_samples_for_sequencing import get_normalization_of_samples
from .post_pcr_bead_purification import get_post_bead_pcr_purification
from cg_lims.models.arnold.base_step import BaseStep


def build_microbial_step_documents(sample_id: str, process_id: str, lims: Lims) -> List[BaseStep]:
    """Building a Step Documents for a Microbial Prep."""

    prep_id = f"{sample_id}_{process_id}"
    step_documents = [
        get_sample_artifact_fields(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_library_prep_nextera(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_buffer_exchange(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_normalization_of_samples(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_normalization_of_mictobial_samples(sample_id=sample_id, lims=lims, prep_id=prep_id),
        get_post_bead_pcr_purification(sample_id=sample_id, lims=lims, prep_id=prep_id),
    ]
    return [document for document in step_documents if document is not None]



from genologics.entities import Artifact
from cg_lims.EPPs.udf.copy.original_well_to_sample import original_well_to_sample
from tests.conftest import server
from cg_lims.exceptions import InvalidValueError
import pytest

def test_assign_position(lims):
    # GIVEN artifacts with container, position and without parent steps.
    server("reception_control_twist")
    artifacts = [Artifact(lims=lims, id="ACC9476A2PA1")]

    # When running function orgwell_to_sample.
    original_well_to_sample(artifacts=artifacts)

    # THEN the corresponding sample will have the matching 'Original Well' and 'Original Container'.
    for artifact in artifacts:
        sample = artifact.samples[0]
        assert sample.udf["Original Well"] == artifact.location[1]
        assert sample.udf["Original Container"] == artifact.location[0].name

def test_replace(lims):
    # GIVEN artifacts with container, position and without parent steps.
    # And sample with already has assigned 'Original Well' and 'Original Container'.
    server("reception_control_twist")
    artifact = [Artifact(lims=lims, id="ACC9476A23PA1")]
    sample = artifact[0].samples[0]
    sample.udf['Original Well'] = "foo"
    sample.udf['Original Container'] = "bar"
    sample.put()

    # When running function original_well_to_sample.
    if artifact[0].location[1] != None and artifact[0].location[0].name != None:
        original_well_to_sample(artifact)
    
    # THEN the corresponding sample will have the matching 'Original Well' and 'Original Container'.
    assert sample.udf["Original Well"] == artifact[0].location[1]
    assert sample.udf["Original Container"] == artifact[0].location[0].name

def test_missing_value(lims):
    # GIVEN that is missing location udf
    server("reception_control_twist")
    artifact = [Artifact(lims=lims, id="ACC9476A3PA1")]
    
    if artifact[0].location[0] != None :
        exit()
    
    # WHEN running function original_well_to_sample
    # THEN MissingUDFsError should be triggered.
    with pytest.raises(InvalidValueError) as error_message:
            original_well_to_sample(artifact)

def test_exceptions_with_parent(lims):
    # GIVEN artifacts with parent steps.
    server("covid_prep")
    artifacts = [Artifact(lims=lims, id="2-1837696"), Artifact(lims=lims, id="2-1837697")]

    for artifact in artifacts:
        if artifact.parent_process == None:
            exit()
    
    # WHEN running function original_well_to_sample.
    # THEN InvalidValueError should be raised.
    with pytest.raises(InvalidValueError) as error_message:
        original_well_to_sample(artifacts=artifacts)

def test_exceptions_pool(lims):
    # GIVEN artifacts with multiple samples, a pool.
    server("missing_reads_pool")
    artifacts = [Artifact(lims=lims, id="2-1439512"), Artifact(lims=lims, id="2-1439513")]

    for artifact in artifacts:
        if len(artifact.samples) == 1:
            exit()
    
    # WHEN running function original_well_to_sample.
    # THEN InvalidValueError should be raised.
    with pytest.raises(InvalidValueError) as error_message:
        original_well_to_sample(artifacts=artifacts)



from typing import Optional

from genologics.lims import Lims
from genologics.entities import Artifact, Sample
from pathlib import Path

import pytest
from click.testing import CliRunner

import threading
import time

from limsmock.server import run_server
from pydantic import BaseModel, Field

from tests.fixtures.flowcell_document import FLOW_CELL_DOCUMENT

PORT = 8000
HOST = "127.0.0.1"


def server(file_name: str):
    """Starting up a server based on file_path"""

    file_path = f"tests/fixtures/{file_name}"
    thread = threading.Thread(
        target=run_server,
        args=(
            file_path,
            HOST,
            PORT,
        ),
    )
    thread.daemon = True
    thread.start()
    time.sleep(0.1)


@pytest.fixture
def lims() -> Lims:
    """Get genologics lims instance"""

    return Lims(f"http://{HOST}:{PORT}", "dummy", "dummy")


@pytest.fixture
def artifact_1(lims) -> Artifact:
    """Basic artifact with id 1. Containing no udfs.
    Related to sample_1."""

    return Artifact(lims, id="1")


@pytest.fixture
def sample_1(lims) -> Sample:
    """Basic sample with id S1. Containing no udfs.
    Related to artifact_1."""

    return Sample(lims, id="S1")


@pytest.fixture
def artifact_2(lims) -> Artifact:
    """Basic artifact with id 2. Containing no udfs.
    Related to sample_2."""

    return Artifact(lims, id="2")


@pytest.fixture
def sample_2(lims) -> Sample:
    """Basic sample with id S2. Containing no udfs.
    Related to artifact_2."""

    return Sample(lims, id="S2")


@pytest.fixture
def config() -> Path:
    """Get file path to config"""

    return Path("tests/fixtures/config.yaml")


@pytest.fixture
def enzymatic_fragmentation_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/Enzymatic_fragmentation"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def kapa_library_preparation_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/KAPA_Library_Preparation"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def kapa_library_preparation_file_missing_udf() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/KAPA_Library_Preparation_missing_udf"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_normalization_csv() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/hamilton_normalization_csv.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_sars_cov2_pooling_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/sars-cov2-hamilton-prep-pooling.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_sars_cov2_indexing_file() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/sars-cov2-hamilton-prep-indexing.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture(name="cli_runner")
def fixture_cli_runner() -> CliRunner:
    """Create a CliRunner"""
    return CliRunner()


@pytest.fixture
def hamilton_buffer_exchange() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/buffer_exchange_hamilton.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def hamilton_buffer_exchange_no_udf() -> str:
    """Get file path to valid json"""

    file_path = "tests/fixtures/buffer_exchange_hamilton_no_udf.txt"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def flow_cell_fixture() -> dict:
    return FLOW_CELL_DOCUMENT



import logging
import sys
import click

from typing import List
from genologics.entities import Process, Artifact
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, InvalidValueError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

def original_well_to_sample(
    artifacts: list
) -> None:
    """Function to copy artifact udf location and set to sample level.

    For each artifact in the artifacts list, copying the location and set
    sample udf 'Original Well' and 'Original Container' to all samples 
    related to the artifact. 

    Arguments:
        artifacts: list of artifacts to copy from"""

    failed = 0
    passed = 0
    for artifact in artifacts:        
        if artifact.parent_process:
            LOG.error(
                f"Error: Not the first step for sample {artifact.id}. Can therefor not get the original container."
            )
            failed += 1
            continue

        elif len(artifact.samples)!=1:
            LOG.error(
                f"Error: more than one sample per artifact for {artifact.id}. Assumes a 1-1 relation between sample and artifact."
            )
            failed += 1
            continue

        sample = artifact.samples[0]
        try:
            sample.udf['Original Well'] = artifact.location[1]
            sample.udf['Original Container'] = artifact.location[0].name
            sample.put()
            passed += 1
        except:
            LOG.error(
                f"Error: Sample {artifact.id} missing udf location. Can therefor not assign values."
        )
            failed += 1

    if failed:
        raise InvalidValueError(
            message=f"Failed to set 'Original Well' and 'Original Container' for {failed} artifacts. Udfs were set on {passed} samples."
        )

@click.command()
@options.input(
    help="Use this flag if you want copy udfs from input artifacts. Default is output artifacts."
)
@click.pass_context
def original_position_to_sample(ctx, input):
    """Script to copy artifact udf to sample udf"""

    process = ctx.obj["process"]
    try:
        artifacts = get_artifacts(process=process, input=input)
        original_well_to_sample(
            artifacts
            )
        message = "Udfs have been set on all samples."
        LOG.info(message)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ProcessUDFs(BaseModel):
    """Define Run Format and Calculate Volumes (Nova Seq)"""

    flowcell_type: Optional[str] = Field(None, alias="FlowCell Type")
    rsb_lot_nr: Optional[str] = Field(None, alias="RSB lot nr")
    final_loading_concentration: Optional[str] = Field(
        None, alias="Final Loading Concentration (pM)"
    )
    total_reads_requested: Optional[str] = Field(
        None, alias="Total nr of Reads Requested (sum of reads to sequence)"
    )
    method: Optional[str] = Field(None, alias="Method")
    protocol_type: Optional[str] = Field(None, alias="Protocol type")


class ArtifactUDFs(BaseModel):
    sequencing_molar_concentration: Optional[float] = Field(None, alias="Concentration (nM)")
    reads_to_sequence: Optional[float] = Field(None, alias="Reads to sequence (M)")
    adjusted_volume_per_sample: Optional[float] = Field(
        None, alias="Adjusted Per Sample Volume (ul)"
    )


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_define_run_format(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Define Run Format and Calculate Volumes (Nova Seq)",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="define_run_format",
        workflow="NovaSeq"
    )



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field
from cg_lims.models.arnold.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ProcessUDFS(BaseModel):
    lot_nr_tagmentation_buffer: Optional[str] = Field(
        None, alias="Lot nr: Tagmentation buffer (TD-buffer)"
    )
    lot_nr_tagmentation_enzyme: Optional[str] = Field(
        None, alias="Lot nr: Tagmentation enzyme (TDE1)"
    )
    lot_nr_index: Optional[str] = Field(None, alias="Lot nr: Index")
    lot_nr_pcr_mix: Optional[str] = Field(None, alias="Lot nr: KAPA HiFi HotStart ReadyMix (2X)")
    pcr_instrument_incubation: Optional[str] = Field(None, alias="PCR instrument incubation")
    pcr_instrument_amplification: Optional[str] = Field(None, alias="PCR instrument amplification")
    nr_pcr_cycles: Optional[int] = Field(None, alias="Nr PCR cycles")
    methods: Optional[str] = Field(None, alias="Methods")
    atlas_version: Optional[str] = Field(None, alias="Atlas Version")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFS

    class Config:
        allow_population_by_field_name = True


def get_library_prep_nextera(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="CG002 - Microbial Library Prep (Nextera)",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFS(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="library_prep",
        workflow="Microbial"
    )



import pytest
from genologics.entities import Sample

from cg_lims.exceptions import MissingUDFsError
from cg_lims.get.udfs import get_udf
from tests.conftest import server


def test_get_udf(sample_1: Sample):
    # GIVEN a sample with a udf "Sequencing Analysis"
    server("flat_tests")
    sample_1.udf["Sequencing Analysis"] = "TESTAPPTAG"

    # WHEN getting the apptag for that sample
    result = get_udf(sample_1, "Sequencing Analysis")

    # THEN the apptag should be returned
    assert result == "TESTAPPTAG"


def test_get_udf_missing_udf(sample_1: Sample):
    # GIVEN a sample with a missing udf "Sequencing Analysis"
    server("flat_tests")
    assert sample_1.udf.get("Sequencing Analysis") is None

    # WHEN getting the apptag for that sample
    with pytest.raises(MissingUDFsError) as error_message:
        get_udf(sample_1, "Sequencing Analysis")

    # THEN the correct exception should be raised
    assert (
        f"UDF Sequencing Analysis not found on sample {sample_1.id}"
        in error_message.value.message
    )



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte

from cg_lims.models.arnold.base_step import BaseStep


class ProcessUDFs(BaseModel):
    pooling_method: Optional[str] = Field(None, alias="Method document (pooling)")
    clean_up_method: Optional[str] = Field(None, alias="Method document (Clean-up)")
    lot_nr_beads_clean_up: Optional[str] = Field(None, alias="Purification beads")
    lot_nr_etoh_clean_up: Optional[str] = Field(None, alias="Ethanol")
    lot_nr_h2o_clean_up: Optional[str] = Field(None, alias="Nuclease-free water")
    lot_nr_resuspension_buffer_clean_up: Optional[str] = Field(None, alias="Resuspension buffer")


class ArtifactUDFs(BaseModel):
    finished_library_concentration: Optional[float] = Field(None, alias="Concentration")
    finished_library_concentration_nm: Optional[float] = Field(None, alias="Concentration (nM)")
    finished_library_size: Optional[int] = Field(None, alias="Size (bp)")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs
    artifact_udfs: ArtifactUDFs

    class Config:
        allow_population_by_field_name = True


def get_pooling_and_cleanup(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Pooling and Clean-up (Cov) v1",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        artifact_udfs=ArtifactUDFs(**analyte.artifact_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="pooling_and_cleanup",
        workflow="sars_cov_2"
    )



from typing import List

from genologics.entities import Process


def get_latest_process(processes: List[Process]) -> Process:
    """Assuming processes of the same process type"""
    processes_to_sort = [(process.date_run, process.id, process) for process in processes]
    processes_to_sort.sort()
    return processes_to_sort[-1][2]



from tests.conftest import server
from pathlib import Path
import pytest

from genologics.entities import Artifact
from cg_lims.EPPs.files.hamilton.normalization_file import get_file_data_and_write
from cg_lims.exceptions import MissingUDFsError


def test_make_hamilton_normalization_file(hamilton_normalization_csv, lims):
    # GIVEN: one Aliquot samples for enzymatic fragmentation TWIST v2 process
    # with 7 input artifacts originating from two 96 well plates and two tubes.
    # All samples have been assigned container barcodes in earlier steps, either from
    # reception control or buffer exchange TWIST.

    server("hamilton_normalization_file")
    file = Path("file_name_1")
    artifacts = [
        Artifact(lims, id="2-2463481"),
        Artifact(lims, id="2-2463482"),
        Artifact(lims, id="2-2463483"),
        Artifact(lims, id="2-2463484"),
        Artifact(lims, id="2-2463485"),
        Artifact(lims, id="2-2463486"),
        Artifact(lims, id="2-2463487"),
    ]

    # WHEN running get_file_data_and_write
    get_file_data_and_write(
        pool=False,
        destination_artifacts=artifacts,
        file=file,
        volume_udf="Sample Volume (ul)",
        buffer_udf="Volume H2O (ul)",
    )

    # THEN the file has been created and the file content is as expected.
    file_content = file.read_text()
    file.unlink()
    assert file_content == hamilton_normalization_csv


def test_hamilton_normalization_file_no_udf(hamilton_normalization_csv, lims):
    # GIVEN: one Aliquot samples for enzymatic fragmentation TWIST v2 process
    # with 7 input artifacts originating from two 96 well plates and two tubes.
    # All samples have been assigned container barcodes in earlier steps, either from
    # reception control or buffer exchange TWIST.

    server("hamilton_normalization_file")
    file = Path("file_name_2")
    incomplete_art = Artifact(lims, id="2-2463481")
    incomplete_art.udf["Volume H2O (ul)"] = None
    incomplete_art.put()
    artifacts = [
        incomplete_art,
        Artifact(lims, id="2-2463482"),
        Artifact(lims, id="2-2463483"),
        Artifact(lims, id="2-2463484"),
        Artifact(lims, id="2-2463485"),
        Artifact(lims, id="2-2463486"),
        Artifact(lims, id="2-2463487"),
    ]

    # WHEN running get_file_data_and_write
    # THEN MissingUDFsError is being raised and no file is created
    with pytest.raises(MissingUDFsError):
        get_file_data_and_write(
            pool=False,
            destination_artifacts=artifacts,
            file=file,
            volume_udf="Sample Volume (ul)",
            buffer_udf="Volume H2O (ul)",
        )
    file.unlink()


def test_hamilton_normalization_file_no_barcode(hamilton_normalization_csv, lims):
    # GIVEN: one Aliquot samples for enzymatic fragmentation TWIST v2 process
    # with 7 input artifacts originating from two 96 well plates and two tubes.
    # All samples have been assigned container barcodes in earlier steps, either from
    # reception control or buffer exchange TWIST.

    server("hamilton_normalization_file")
    file = Path("file_name_3")
    incomplete_art = Artifact(lims, id="2-2463481")
    incomplete_art.udf["Output Container Barcode"] = None
    incomplete_art.put()
    artifacts = [
        incomplete_art,
        Artifact(lims, id="2-2463482"),
        Artifact(lims, id="2-2463483"),
        Artifact(lims, id="2-2463484"),
        Artifact(lims, id="2-2463485"),
        Artifact(lims, id="2-2463486"),
        Artifact(lims, id="2-2463487"),
    ]

    # WHEN running get_file_data_and_write
    # THEN MissingUDFsError is being raised and no file is created
    with pytest.raises(MissingUDFsError):
        get_file_data_and_write(
            pool=False,
            destination_artifacts=artifacts,
            file=file,
            volume_udf="Sample Volume (ul)",
            buffer_udf="Volume H2O (ul)",
        )
    file.unlink()



import pytest
from genologics.entities import Artifact, Process

from cg_lims.EPPs.udf.calculate.calculate_average_size_and_set_qc import calculate_average_size, set_average_and_qc
from cg_lims.exceptions import MissingUDFsError, MissingCgFieldError
from cg_lims.set.udfs import copy_artifact_to_artifact
from tests.conftest import server


def test_calculate_average_size(lims):
    # GIVEN a list of artifacts where some have udf "size (bp)"
    # set to some integer and some are None.
    # And given that the average of the values is 244

    server("calculate_average_size_and_set_qc")
    process = Process(lims, id="24-297378")
    all_artifacts = process.all_inputs()

    # WHEN running calculate_average_size
    average = calculate_average_size(all_artifacts=all_artifacts)

    # THEN the calculated average is 244
    assert int(average) == 244


def test_calculate_average_size_no_sizes_given(lims):
    # GIVEN a list of artifacts that dont have the udf "size (bp)" set.

    server("calculate_average_size_and_set_qc")
    process = Process(lims, id="24-297378")
    all_artifacts = process.all_inputs()
    for artifact in all_artifacts:
        artifact.udf.clear()
        artifact.put()

    # WHEN running calculate_average_size
    # THEN MissingUDFsError is being raised
    with pytest.raises(MissingUDFsError) as error_message:
        average = calculate_average_size(all_artifacts=all_artifacts)


def test_calculate_average_size_ntc_only(lims):
    # Given a list of NTC artifacts that have "Size (bp)" set.

    server("calculate_average_size_and_set_qc")
    process = Process(lims, id="24-297378")
    all_artifacts = process.all_inputs()
    ntc_artifacts = [artifact for artifact in all_artifacts if artifact.name[0:3] == "NTC"]

    # WHEN running calculate_average_size
    # THEN MissingUDFsError is being raised
    with pytest.raises(MissingUDFsError) as error_message:
        average = calculate_average_size(all_artifacts=ntc_artifacts)


def test_set_average_and_qc(lims):
    # GIVEN a list of artifacts with mixed QC flags and an average of 400, which is over the lower threshold
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id="24-297378")
    all_artifacts = process.all_inputs()
    all_artifacts[0].qc_flag = "FAILED"
    all_artifacts[1].qc_flag = "PASSED"

    # WHEN running set_average_and_qc
    set_average_and_qc(average=400, lower_threshold="300", upper_threshold="", all_artifacts=all_artifacts)

    # THEN the average set is 400 and the QC flags of the artifacts are unchanged if already existing,
    # or set to "PASSED" if previously set to None
    for artifact in all_artifacts:
        assert artifact.udf["Average Size (bp)"] == "400"
    assert all_artifacts[0].qc_flag == "FAILED"
    for artifact in all_artifacts[1:]:
        assert artifact.qc_flag == "PASSED"


def test_set_average_and_qc_under_lt(lims):
    # GIVEN a list of artifacts with mixed QC flags and an average of 200, which is under the lower threshold
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id="24-297378")
    all_artifacts = process.all_inputs()
    all_artifacts[0].qc_flag = "FAILED"
    all_artifacts[1].qc_flag = "PASSED"

    # WHEN running set_average_and_qc
    set_average_and_qc(average=200, lower_threshold="300", upper_threshold="600", all_artifacts=all_artifacts)

    # THEN the average set is 200 and the QC flags of the artifacts are all set to "FAILED",
    # no matter what the previous state was
    for artifact in all_artifacts:
        assert artifact.udf["Average Size (bp)"] == "200"
        assert artifact.qc_flag == "FAILED"


def test_set_average_and_qc_over_ut(lims):
    # GIVEN a list of artifacts with mixed QC flags and an average of 700, which is over the upper threshold
    server("calculate_average_size_and_set_qc")
    process = Process(lims, id="24-297378")
    all_artifacts = process.all_inputs()
    all_artifacts[0].qc_flag = "FAILED"
    all_artifacts[1].qc_flag = "PASSED"

    # WHEN running set_average_and_qc
    set_average_and_qc(average=700, lower_threshold="300", upper_threshold="600", all_artifacts=all_artifacts)

    # THEN the average set is 700 and the QC flags of the artifacts are all set to "FAILED",
    # no matter what the previous state was
    for artifact in all_artifacts:
        assert artifact.udf["Average Size (bp)"] == "700"
        assert artifact.qc_flag == "FAILED"



from genologics.entities import Process

from cg_lims.EPPs.arnold.prep import build_step_documents
from cg_lims.models.arnold.base_step import BaseStep

from tests.conftest import (
    server,
)


COVID_STEP_TYPES = {"library_prep", "reception_control", "pooling_and_cleanup"}


def test_covid_prep(lims):
    # GIVEN: A lims with a process: "24-240270" (Aggregate QC (Library Validation) (Cov)).
    # Where the samples in the process has gone through the whole covid prep workflow:

    server("covid_prep")
    process = Process(lims=lims, id="24-240270")

    # WHEN running build_step_documents for the covid prep
    step_documents = build_step_documents(lims=lims, process=process, prep_type="cov")

    # THEN assert BaseStep documents are created and all requiered step types in the covid prep workflow are represented
    for document in step_documents:
        assert isinstance(document, BaseStep)
    assert {document.step_type for document in step_documents} == COVID_STEP_TYPES



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.objects import BaseAnalyte
from cg_lims.models.arnold.base_step import BaseStep


class ProcessUDFS(BaseModel):
    lot_nr_covaris_tube: Optional[str] = Field(None, alias="Lot no: Covaris plate or tube")
    fragmentation_method: Optional[str] = Field(None, alias="Method Document 1")
    covaris_protocol: Optional[str] = Field(None, alias="Protocol")
    lot_nr_resuspension_buffer_fragmentation: Optional[str] = Field(
        None, alias="Lot no: Resuspension Buffer"
    )
    methods: Optional[str] = Field(None, alias="Methods")
    atlas_version: Optional[str] = Field(None, alias="Atlas Version")
    lot_nr_ethanol: Optional[str] = Field(None, alias= "Lot no: Ethanol")
    lot_nr_water: Optional[str] = Field(None, alias= "Lot no: Nuclease free water")
    lot_nr_beads: Optional[str] = Field(None, alias= "Lot no: Beads")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFS

    class Config:
        allow_population_by_field_name = True


def get_fragemnt_dna_truseq(lims: Lims, sample_id: str, prep_id: str) -> ArnoldStep:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Fragment DNA (TruSeq DNA)",
    )

    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFS(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="fragment_dna",
        workflow="WGS",
    )



import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import FailingQCError, LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts
from cg_lims.get.samples import get_one_sample_from_artifact

LOG = logging.getLogger(__name__)


def get_qc(source: str, conc: float, amount: float) -> str:
    """QC-criteria depends on sample source, total amount and sample concentration. See AMS doc 1117, 1993 and 2125.
    The volume is subtracted by 3 in the calculations. This is beacause the lab uses 3 ul in the initial qc measurements.
    """

    qc = "FAILED"

    if source == "cfDNA":
        if amount >= 10 and conc <= 250 and conc >= 0.2:
            qc = "PASSED"
    else:
        if amount >= 250 and conc <= 250 and conc >= 8.33:
            qc = "PASSED"

    return qc


def calculate_amount_and_set_qc(artifacts: List[Artifact]) -> None:
    """Calculate amount and set qc for twist samples"""

    missing_udfs_count = 0
    qc_fail_count = 0
    for artifact in artifacts:
        sample = get_one_sample_from_artifact(artifact)
        source = sample.udf.get("Source")
        vol = artifact.udf.get("Volume (ul)")
        conc = artifact.udf.get("Concentration")
        if None in [source, conc, vol]:
            missing_udfs_count += 1
            continue

        amount = conc * (vol - 3)
        artifact.udf["Amount (ng)"] = amount
        qc = get_qc(source, conc, amount)
        if qc == "FAILED":
            qc_fail_count += 1
            LOG.info(
                f"Sample {sample.id} failed qc. Source: {source} Amount: {amount} Concentration: {conc}"
            )
        artifact.qc_flag = qc
        artifact.put()

    if missing_udfs_count:
        raise MissingUDFsError(f"Udf missing for {missing_udfs_count} sample(s).")
    if qc_fail_count:
        raise FailingQCError(
            f"Amounts have been calculated and qc flags set for all samples. QC failed for {qc_fail_count} sample(s)."
        )


@click.command()
@click.pass_context
def twist_qc_amount(ctx):
    """Calculates amount of samples for qc validation."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        artifacts = get_artifacts(process=process, measurement=True)
        calculate_amount_and_set_qc(artifacts)
        message = "Amounts have been calculated and qc flags set for all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



from genologics.entities import Process

from cg_lims.EPPs.arnold.prep import build_step_documents
from cg_lims.models.arnold.base_step import BaseStep

from tests.conftest import (
    server,
)


TWIST_STEP_TYPES = {
    "enzymatic_fragmentation",
    "buffer_exchange",
    "kapa_library_preparation",
    "pool_samples",
    "bead_purification",
    "aliquot_samples_for_enzymatic_fragmentation",
    "hybridize_library",
    "capture_and_wash",
    "amplify_captured_library",
    "reception_control",
    "normalization_for_sequencing",
}


def test_twist_prep(lims):
    # GIVEN: A lims with a process: "24-309487" (Aggregate QC (Library Validation) TWIST).
    # Where the samples in the process has gone through the whole twist prep workflow:

    server("twist_prep")
    process = Process(lims=lims, id="24-309487")

    # WHEN running build_step_documents
    step_documents = build_step_documents(lims=lims, process=process, prep_type="twist")
    # THEN assert BaseStep documents are created and all step types in the twist workflow are represented
    for document in step_documents:
        assert isinstance(document, BaseStep)
    assert {document.step_type for document in step_documents} == TWIST_STEP_TYPES



from datetime import date
from typing import Optional
from genologics.entities import Entity
from genologics.lims import Lims

from cg_lims.exceptions import MissingUDFsError
import logging

LOG = logging.getLogger(__name__)


def get_udf_type(lims: Lims, udf_name: str, attach_to_name: str) -> Optional:
    """Get udf type.

    Args:
        udf_name:
            eg: 'Concentration (nM)', 'Comment',...
        attach_to_name: name of entity to which the udf is attached
            eg. 'ResultFile', 'Analyte', 'Aggregate QC (RNA)',...
    """
    udf_types = {"String": str, "Numeric": float, "Date": date}

    udf_configs = lims.get_udfs(name=udf_name, attach_to_name=attach_to_name)
    udf_configs[0].get()
    udf_type = udf_configs[0].root.attrib["type"]

    return udf_types[udf_type]


def get_udf(entity: Entity, udf: str) -> str:
    """Returns the value of a udf on a entity.
    entity can be a Artifact, Sample, Process, etc"""
    try:
        return entity.udf[udf]
    except Exception:
        message = f"UDF {udf} not found on {entity._TAG} {entity.id}!"
        LOG.error(message)
        raise MissingUDFsError(message)



"""Unit tests for cg_lims.EPPs.udf.calculate.maf_calculate_volume"""
import logging

import pytest
from genologics.entities import Artifact
from genologics.lims import Lims
from pydantic import ValidationError

from cg_lims.EPPs.udf.calculate.maf_calculate_volume import (
    QC_FAILED,
    QC_PASSED,
    MafVolumes,
    calculate_volume,
)
from tests.conftest import server

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "sample_concentration, final_volume, water_volume, sample_volume, qc_flag",
    [
        (19, 15.0, 11.842105263157894, 3.1578947368421053, QC_FAILED),
        (20, 15.0, 12.0, 3, QC_PASSED),
        (21, 15.75, 12.75, 3, QC_PASSED),
        (239, 179.25, 176.25, 3, QC_PASSED),
        (243, 182.25, 179.25, 3, QC_PASSED),
        (244, 122.0, 120.0, 2, QC_FAILED),
        (363, 181.5, 179.5, 2, QC_FAILED),
        (364, 91.0, 90.0, 1, QC_FAILED),
        (723, 180.75, 179.75, 1, QC_FAILED),
        (724, 90.5, 90.0, 0.5, QC_FAILED),
        (1443, 180.375, 179.875, 0.5, QC_FAILED),
    ],
)
def test_maf_volume_model(sample_concentration, sample_volume, final_volume, water_volume, qc_flag):
    # GIVEN a pydantic model for MAF volumes

    # WHEN calculating the volumes in the validators with valid values for 'sample_concentration'
    result = MafVolumes(sample_concentration=sample_concentration)

    # THEN the correct results should be set
    assert result.sample_concentration == sample_concentration
    assert result.final_volume == final_volume
    assert result.water_volume == water_volume
    assert result.sample_volume == sample_volume
    assert result.qc_flag == qc_flag


@pytest.mark.parametrize(
    "sample_concentration",
    [
        3.99,
    ],
)
def test_maf_volume_model_sample_concentration_value_error(sample_concentration):
    # GIVEN a pydantic model for MAF volumes

    # WHEN calculating the volumes in the validators with an invalid value for
    # 'sample_concentration'
    with pytest.raises(ValueError) as error_message:
        MafVolumes(sample_concentration=sample_concentration)

    # THEN the validator for sample_concentration should raise a ValueError exception
    assert error_message.value.errors()[0]["loc"][0] == "sample_concentration"
    assert error_message.value.errors()[0]["type"] == "value_error"
    assert error_message.value.errors()[0]["msg"] == "Too low or missing concentration"


@pytest.mark.parametrize(
    "sample_concentration",
    [
        None,
    ],
)
def test_maf_volume_model_sample_concentration_validation_error(sample_concentration):
    # GIVEN a pydantic model for MAF volumes

    # WHEN calculating the volumes in the validator when there is no 'sample_concentration'
    with pytest.raises(ValidationError) as error_message:
        MafVolumes(sample_concentration=sample_concentration)

    # THEN pydantic should raise a ValidationError exception
    assert error_message.value.errors()[0]["loc"][0] == "sample_concentration"
    assert error_message.value.errors()[0]["type"] == "type_error.none.not_allowed"
    assert error_message.value.errors()[0]["msg"] == "none is not an allowed value"


@pytest.mark.parametrize(
    "sample_concentration",
    [
        9001,
    ],
)
def test_maf_volume_model_sample_volume_value_error(sample_concentration):
    # GIVEN a pydantic model for MAF volumes

    # WHEN calculating the volumes in the validator when there is no 'sample_concentration'
    with pytest.raises(ValueError) as error_message:
        MafVolumes(sample_concentration=sample_concentration)

    # THEN the validator for sample_volume should raise a ValueError exception
    assert error_message.value.errors()[0]["loc"][0] == "sample_volume"
    assert error_message.value.errors()[0]["type"] == "value_error"
    assert error_message.value.errors()[0]["msg"] == "Sample concentration not in valid range"


@pytest.mark.parametrize(
    "sample_concentration, final_volume, water_volume, sample_volume, qc_flag",
    [
        (243, 182.25, 179.25, 3, QC_PASSED),
    ],
)
def test_calculate_volume(
    sample_concentration,
    sample_volume,
    final_volume,
    water_volume,
    qc_flag,
    artifact_1: Artifact,
):
    # GIVEN a sample with a valid sample concentration:
    server("flat_tests")
    artifact_1.udf["Concentration"] = sample_concentration
    artifacts = [artifact_1]

    # WHEN calculating the volumes
    calculate_volume(artifacts)

    # THEN the sample udfs should be set to the correct values
    artifact_1.qc_flag = qc_flag
    artifact_1.udf["Final Volume (uL)"] = final_volume
    artifact_1.udf["Volume H2O (ul)"] = water_volume
    artifact_1.udf["Volume of sample (ul)"] = sample_volume


@pytest.mark.parametrize(
    "sample_concentration",
    [
        3.99,
    ],
)
def test_calculate_volume_exception(
    sample_concentration,
    artifact_1: Artifact,
    caplog,
):
    # GIVEN a sample with a sample concentration that is below the required final concentration
    server("flat_tests")
    artifact_1.udf["Concentration"] = sample_concentration
    artifacts = [artifact_1]

    # WHEN calculating the volumes
    with pytest.raises(Exception) as error_message:
        calculate_volume(artifacts)

    # THEN an Exception should be raised
    assert "Too low or missing concentration" in caplog.text
    assert "MAF volume calculations failed for" in error_message.value.message



"""Base cg_lims calss definitions
"""
from typing import Optional, List

from genologics.entities import Artifact, Sample, Process
from genologics.lims import Lims

from cg_lims.exceptions import (
    MissingUDFsError,
    ZeroReadsError,
    MissingArtifactError,
    MissingProcessError,
)
from cg_lims.get.artifacts import (
    get_sample_artifact,
    get_latest_analyte,
)
import logging
from statistics import mean

LOG = logging.getLogger(__name__)


class Pool:
    """Collecting informatiuon general to any pool"""

    def __init__(self, pool_artifact: Artifact):
        self.total_reads_missing = 0
        self.artifact = pool_artifact

    def get_total_reads_missing(self) -> None:
        """Get the total number of missing reads in the pool"""

        for art in self.artifact.input_artifact_list():
            reads = art.samples[0].udf.get("Reads missing (M)")
            if reads is None:
                raise MissingUDFsError("Missing udfs: Reads missing (M)")
            self.total_reads_missing += reads
        if self.total_reads_missing == 0:
            raise ZeroReadsError("All samples seem to have Missing Reads = 0!")


class BaseAnalyte:
    """This class defines a Analyte by:

    sample_id: The submitted sample from which the analyte is derived.
    process_types: The process in which the analyte was generated (the artifact.parent_process).
                  If process_types is None, then the analyte is the original submitted sample.
    """

    def __init__(
        self,
        lims: Lims,
        sample_id: str,
        process_type: str = None,
        optional_step: bool = False,
    ):
        self.lims: Lims = lims
        self.sample_id: str = sample_id
        self.process_type: str = process_type
        self.optional_step: bool = optional_step
        self.artifact: Optional[Artifact] = self.get_artifact()
        self.process = self.get_process()

    def get_process(self) -> Optional[Process]:
        """Returning parent process if existing"""
        try:
            return self.artifact.parent_process
        except:
            return None

    def get_well(self) -> Optional[str]:
        """Returning artifact well if existing."""
        if not (self.artifact and self.artifact.location and self.artifact.location[1] != "1:1"):
            return None
        return self.artifact.location[1]

    def get_container_name(self) -> Optional[str]:
        """Returning artifact container name if existing."""
        if not (self.artifact and self.artifact.container and self.artifact.container.name):
            return None
        return self.artifact.container.name

    def get_container_id(self) -> Optional[str]:
        """Returning artifact container id if existing."""
        if not (self.artifact and self.artifact.container and self.artifact.container.id):
            return None
        return self.artifact.container.id

    def get_container_type(self) -> Optional[str]:
        """Returning artifact container type if existing."""
        if not (self.artifact and self.artifact.container and self.artifact.container.type):
            return None
        return self.artifact.container.type.name

    def get_index(self) -> Optional[str]:
        """Returning artifact index if existing."""
        if not (
            self.artifact
            and self.artifact.reagent_labels
            and len(self.artifact.reagent_labels) == 1
        ):
            return None
        return self.artifact.reagent_labels[0]

    def get_nr_samples(self) -> Optional[int]:
        """Returning nr samples in pool if existing."""
        if not (self.artifact and self.artifact.samples and len(self.artifact.samples) > 1):
            return None
        return len(self.artifact.samples)

    def get_artifact_name(self) -> Optional[int]:
        """Returning artifact name if existing."""
        if not self.artifact:
            return None
        return self.artifact.name

    def get_artifact(self) -> Optional[Artifact]:
        """Getting the analyte artifact based on sample_id and process_types.
        If process_types is None, then the analyte is the original submitted sample."""
        if not self.process_type:
            sample = Sample(self.lims, id=self.sample_id)
            return get_sample_artifact(sample=sample, lims=self.lims)
        try:
            return get_latest_analyte(
                lims=self.lims, sample_id=self.sample_id, process_types=[self.process_type]
            )
        except MissingArtifactError as e:
            LOG.info(e.message)
            if self.optional_step:
                return None
            raise e

    def get_date(self):
        """Date when artifact was produced."""
        if not (self.process and self.process.date_run):
            return None

        return self.process.date_run

    def process_udfs(self) -> dict:
        """Filtering process udfs by process_udf_model."""
        if self.process:
            return dict(self.process.udf.items())

        if self.optional_step:
            return {}
        raise MissingProcessError(
            message=f"sample didnt pass through requiered process {self.process_type}"
        )

    def artifact_udfs(self) -> dict:
        """Filtering artifact udfs by artifact_udf_model."""
        if self.artifact:
            return dict(self.artifact.udf.items())
        if self.optional_step:
            return {}
        raise MissingProcessError(
            message=f"sample didnt pass through requiered process {self.process_type}"
        )

    def base_fields(self) -> dict:

        return dict(
            well_position=self.get_well(),
            container_name=self.get_container_name(),
            container_id=self.get_container_id(),
            container_type=self.get_container_type(),
            index_name=self.get_index(),
            nr_samples_in_pool=self.get_nr_samples(),
            artifact_name=self.get_artifact_name(),
            lims_step_name=self.process_type,
            date_run=self.get_date(),
        )



import pytest
from genologics.entities import Process

from cg_lims.EPPs.udf.set.set_method import get_path
from cg_lims.exceptions import AtlasResponseFailedError
from tests.conftest import (
    server,
)

import responses
import requests
import logging


@pytest.fixture()
@responses.activate
def atlas_response_mock():
    atlas_document_path = "atlas/document/path.md"
    url = "https://atlas/api/SomeTitle/path"
    responses.add(responses.GET, url, json=atlas_document_path, status=200)
    return requests.get(url)


def test_get_path(lims, mocker, atlas_response_mock):
    # GIVEN: a lims with a process: id="24-196211"
    # with a udf named "Method Document" with a value of some method document title in atlas: "SomeTitle"  and
    # GIVEN: a atlas api with a endpoint "https://atlas/api/SomeTitle/path" returning "atlas/document/path.md"

    server("flat_tests")
    document_udf = "Method Document"
    process = Process(lims=lims, id="24-196211")
    process.udf[document_udf] = "SomeTitle"
    process.put()
    mocker.patch.object(requests, "get")
    requests.get.return_value = atlas_response_mock

    # WHEN running get_path with the process, the atlas api host and the document udf
    document_path = get_path(
        process=process, document_title="SomeTitle", atlas_host="https://atlas/api"
    )

    # THEN assert get_path returns the document path
    assert document_path == "atlas/document/path.md"


@pytest.fixture()
@responses.activate
def atlas_response_mock_fail():
    url = "https://atlas/api/SomeTitle/path"
    responses.add(responses.GET, url, status=404, json={"detail": "document not found"})
    return requests.get(url)


def test_get_path_wrong_document_title(lims, mocker, atlas_response_mock_fail):
    # GIVEN: a lims with a process: id="24-196211"
    # with a udf named "Method Document" with a value of some method document title that dont exist in atlas: "SomeWrongTitle"  and
    # GIVEN: a atlas api with a endpoint "https://atlas/api/SomeTitle/path" returning "atlas/document/path.md"

    server("flat_tests")
    document_udf = "Method Document"
    process = Process(lims=lims, id="24-196211")
    process.udf[document_udf] = "SomeWrongTitle"
    process.put()
    mocker.patch.object(requests, "get")
    requests.get.return_value = atlas_response_mock_fail

    # WHEN running get_path with the process, the atlas api host and the document udf.
    # THEN: assert AtlasResponseFailedError is being raised
    with pytest.raises(AtlasResponseFailedError) as error:
        document_path = get_path(
            process=process, document_title="SomeWrongTitle", atlas_host="https://atlas/api"
        )

    # THEN: assert the error message is fetched from the api response
    assert error.value.message == '404 : {"detail": "document not found"}'



import logging
import sys
from typing import Tuple

import click

from cg_lims.exceptions import LimsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def sum_reads_in_pool(artifacts: list) -> Tuple[int, int]:
    """Summing the missing reads for all samples in one pool."""

    failed_arts = 0
    passed_arts = 0

    for artifact in artifacts:
        if len(artifact.samples) == 1:
            continue

        missing_reads_pool = []
        for sample in artifact.samples:
            if sample.udf.get("Reads missing (M)") is None:
                failed_arts += 1
                missing_reads_pool = []
                break
            missing_reads_pool.append(sample.udf["Reads missing (M)"])
        if missing_reads_pool:
            artifact.udf["Missing reads Pool (M)"] = sum(missing_reads_pool)
            artifact.put()
            passed_arts += 1
    return passed_arts, failed_arts


@click.command()
@click.pass_context
def missing_reads_in_pool(ctx):
    """Script to calculate missing reads in a pool"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]

    try:
        artifacts = get_artifacts(process=process, input=False)
        passed_arts, failed_arts = sum_reads_in_pool(artifacts=artifacts)
        message = f"Updated {passed_arts}. Ignored {failed_arts} due to missing Sample UDFs: 'Reads missing (M)'"
        if failed_arts:
            LOG.error(message)
            sys.exit(message)
        else:
            LOG.info(message)
            click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



#!/usr/bin/env python

import click

from cg_lims.EPPs.udf.calculate.calculate_amount_ng import calculate_amount_ng
from cg_lims.EPPs.udf.calculate.calculate_beads import calculate_beads
from cg_lims.EPPs.udf.calculate.calculate_microbial_aliquot_volumes import (
    calculate_microbial_aliquot_volumes,
)
from cg_lims.EPPs.udf.calculate.calculate_resuspension_buffer_volumes import (
    calculate_resuspension_buffer_volume,
)
from cg_lims.EPPs.udf.calculate.calculate_water import volume_water
from cg_lims.EPPs.udf.calculate.calculate_water_volume_rna import (
    calculate_water_volume_rna,
)
from cg_lims.EPPs.udf.calculate.get_missing_reads import get_missing_reads
from cg_lims.EPPs.udf.calculate.maf_calculate_volume import maf_calculate_volume
from cg_lims.EPPs.udf.calculate.molar_concentration import molar_concentration
from cg_lims.EPPs.udf.calculate.sum_missing_reads_in_pool import missing_reads_in_pool
from cg_lims.EPPs.udf.calculate.twist_aliquot_amount import twist_aliquot_amount
from cg_lims.EPPs.udf.calculate.twist_aliquot_volume import twist_aliquot_volume
from cg_lims.EPPs.udf.calculate.twist_get_volumes_from_buffer import (
    get_volumes_from_buffer,
)
from cg_lims.EPPs.udf.calculate.calculate_average_size_and_set_qc import calculate_average_size_and_set_qc

# commands
from cg_lims.EPPs.udf.calculate.twist_pool import twist_pool
from cg_lims.EPPs.udf.calculate.twist_qc_amount import twist_qc_amount


@click.group(invoke_without_command=True)
@click.pass_context
def calculate(ctx):
    """Main entry point of calculate commands"""
    pass


calculate.add_command(twist_pool)
calculate.add_command(twist_aliquot_amount)
calculate.add_command(twist_aliquot_volume)
calculate.add_command(twist_qc_amount)
calculate.add_command(get_volumes_from_buffer)
calculate.add_command(get_missing_reads)
calculate.add_command(calculate_amount_ng)
calculate.add_command(volume_water)
calculate.add_command(molar_concentration)
calculate.add_command(calculate_beads)
calculate.add_command(missing_reads_in_pool)
calculate.add_command(maf_calculate_volume)
calculate.add_command(calculate_resuspension_buffer_volume)
calculate.add_command(calculate_water_volume_rna)
calculate.add_command(calculate_microbial_aliquot_volumes)
calculate.add_command(calculate_average_size_and_set_qc)



import pytest
from genologics.entities import Artifact, Process

from cg_lims.EPPs.udf.copy.aggregate_qc_flags_and_copy_fields import get_copy_tasks, copy_udfs
from cg_lims.get.artifacts import get_artifacts
from cg_lims.exceptions import MissingUDFsError
from tests.conftest import server


def test_get_copy_tasks_3_tasks(lims):
    # GIVEN a process with 3 copy task udfs:
    # Copy task 1 - Source Field -> Size (bp)
    # Copy task 1 - Source Step -> Tapestation QC TWIST v2
    # Copy task 2 - Source Field -> Concentration
    # Copy task 2 - Source Step -> Qubit QC (Library Validation) v2
    # Copy task 3 - Source Field -> Amount (ng)
    # Copy task 3 - Source Step -> Qubit QC (Library Validation) v2
    server("aggregate_qc")
    process = Process(lims, id="24-306120")

    # WHEN running get_copy_tasks
    source_udfs = get_copy_tasks(process=process)

    # THEN the correct source udfs are collected
    correct_source_udfs = {
        "Copy task 1": {"Source Field": "Size (bp)", "Source Step": "Tapestation QC TWIST v2"},
        "Copy task 2": {
            "Source Field": "Concentration",
            "Source Step": "Quantit QC (Library Validation) TWIST v2",
        },
        "Copy task 3": {
            "Source Field": "Amount (ng)",
            "Source Step": "Quantit QC (Library Validation) TWIST v2",
        },
    }
    assert source_udfs == correct_source_udfs


def test_get_copy_tasks_2_tasks(lims):
    # GIVEN a process with 2 copy task udfs:
    # Copy task 1 - Source Field -> Concentration (nM)
    # Copy task 1 - Source Step -> CG002 - qPCR QC (Library Validation)
    # Copy task 2 - Source Field -> Size (bp)
    # Copy task 2 - Source Step -> CG002 - qPCR QC (Library Validation)
    server("wgs_prep")
    process = Process(lims, id="24-240276")

    # WHEN running get_copy_tasks
    source_udfs = get_copy_tasks(process=process)

    # THEN the correct source udfs are collected
    correct_source_udfs = {
        "Copy task 1": {
            "Source Field": "Concentration (nM)",
            "Source Step": "CG002 - qPCR QC (Library Validation)",
        },
        "Copy task 2": {
            "Source Field": "Size (bp)",
            "Source Step": "CG002 - qPCR QC (Library Validation)",
        },
    }
    assert source_udfs == correct_source_udfs


def test_get_copy_tasks_no_tasks(lims):
    # GIVEN a process with no copy task udfs
    server("wgs_prep")
    process = Process(lims, id="24-240274")

    # WHEN running get_copy_tasks
    # THEN MissingUDFsError is raised
    with pytest.raises(MissingUDFsError) as error_message:
        source_udfs = get_copy_tasks(process=process)


def test_get_copy_tasks_wrong_type(lims):
    # GIVEN a process with no copy task udfs
    server("wgs_prep")
    process = Process(lims, id="24-240274")
    process.udf["Copy task 1 - Source Well"] = "1:1"

    # WHEN running get_copy_tasks
    # THEN MissingUDFsError is raised
    with pytest.raises(MissingUDFsError) as error_message:
        source_udfs = get_copy_tasks(process=process)



from typing import List, Optional, Literal

from genologics.entities import Artifact, Process, Sample
from genologics.lims import Lims

from cg_lims.exceptions import FileError, MissingArtifactError
from datetime import datetime
import logging

LOG = logging.getLogger(__name__)


def get_sample_artifact(lims: Lims, sample: Sample) -> Artifact:
    """Returning the initial artifact related to a sample.
    Assuming first artifact is allways named sample.id + 'PA1."""
    sample_artifact_id = f"{sample.id}PA1"
    artifact = Artifact(lims, id=sample_artifact_id)
    try:
        artifact.get()
    except:
        message = f"no artifact with id {sample_artifact_id}"
        LOG.error(message)
        raise MissingArtifactError(message=message)
    return artifact


def get_output_artifacts(
    lims: Lims,
    process: Process,
    output_type: Literal["ResultFile", "Analyte"],
    output_generation_type: List[Literal["PerInput", "PerReagentLabel", "PerAllInputs"]],
) -> List[Artifact]:
    """Get output 'artifacts' based on output_generation_type and output_type"""

    artifacts = [
        Artifact(lims, id=output["limsid"])
        for input, output in process.input_output_maps
        if output["output-generation-type"] in output_generation_type
        and output["output-type"] == output_type
    ]
    return list(frozenset(artifacts))


def get_artifacts(
    process: Process,
    input: Optional[bool] = False,
    measurement: Optional[bool] = False,
    reagent_label: Optional[bool] = False,
) -> List[Artifact]:
    """If inputs is True, returning all input analytes of the process,
    otherwise returning all output analytes of the process"""

    if input:
        return process.all_inputs(unique=True)
    elif measurement:
        return get_output_artifacts(
            lims=process.lims,
            process=process,
            output_type="ResultFile",
            output_generation_type=["PerInput"],
        )
    elif reagent_label:
        return get_output_artifacts(
            lims=process.lims,
            process=process,
            output_type="ResultFile",
            output_generation_type=["PerReagentLabel"],
        )
    else:
        return get_output_artifacts(
            lims=process.lims,
            process=process,
            output_type="Analyte",
            output_generation_type=["PerInput", "PerAllInputs"],
        )


def get_artifact_by_name(process: Process, name: str) -> Artifact:
    """Searches for output artifacts with name <file_name>
    Raises error if more than one artifact with same name!"""

    artifacts = list(filter(lambda a: a.name in [name], process.all_outputs()))

    if len(artifacts) > 1:
        message = f"more than one output artifact named {name}"
        LOG.error(message)
        raise FileError(message=message)

    return artifacts[0]


def filter_artifacts(artifacts: List[Artifact], udf: str, value) -> List[Artifact]:
    """Returning a list of only artifacts with udf==value"""

    if udf is None:
        return artifacts
    else:
        return [a for a in artifacts if a.udf.get(udf) == value]


def get_latest_artifact(lims_artifacts: List[Artifact]) -> Artifact:
    """Returning the latest generated artifact in the list of artifacts"""
    if not lims_artifacts:
        message = "Can not find latest artifact from a list of no artifacts"
        LOG.error(message)
        raise MissingArtifactError(message=message)

    artifacts = []
    for artifact in lims_artifacts:
        date = artifact.parent_process.date_run or datetime.today().strftime("%Y-%m-%d")
        artifacts.append((date, artifact.id, artifact))

    artifacts.sort()
    date, id, latest_art = artifacts[-1]
    return latest_art


def get_latest_analyte(
    lims: Lims, sample_id: str, process_types: Optional[List[str]], sample_artifact: bool = False
) -> Artifact:
    """Getting the most recently generated analyte by process_types and sample_id.

    Searching for all Analytes associated with <sample_id> that were produced by <process_types>.

    Returning the analyte with latest parent_process.date_run.
    If there are many such artifacts only one will be returned."""

    if sample_artifact and not process_types:
        return get_sample_artifact(lims=lims, sample=Sample(lims, sample_id))

    lims_artifacts = lims.get_artifacts(
        samplelimsid=sample_id,
        type="Analyte",
        process_type=process_types,
    )

    if sample_artifact and not lims_artifacts:
        return get_sample_artifact(lims=lims, sample=Sample(lims, sample_id))

    if not lims_artifacts:
        message = f"Could not find a artifact with sample: {sample_id} generated by process: {' ,'.join(process_types)}"
        LOG.error(message)
        raise MissingArtifactError(message=message)

    return get_latest_artifact(lims_artifacts=lims_artifacts)


def get_latest_result_files(
    lims: Lims,
    sample_id: str,
    process_types: Optional[List[str]],
    output_generation_type: Literal["PerInput", "PerReagentLabel"],
) -> List[Artifact]:
    """This function will make a lot of queries if the nr of artifacts in the step are many and the artifacts are pools.
    (At least 3+2n+nk queries to teh database, where n is the number of input artifacts and k is the number of samples
    per artifact if artifact is pool.)

    Only use this function if you have to."""

    all_result_files = lims.get_artifacts(
        samplelimsid=sample_id,
        type="ResultFile",
        process_type=process_types,
    )
    sample = Sample(lims=lims, id=sample_id)
    latest = get_latest_artifact(lims_artifacts=all_result_files)
    latest_process: Process = latest.parent_process

    specific_result_files = []
    for input, output in latest_process.input_output_maps:
        if output["output-generation-type"] != output_generation_type:
            continue
        artifact = Artifact(lims, id=output["limsid"])
        if sample not in artifact.samples:
            continue

        specific_result_files.append(artifact)

    if not specific_result_files:
        message = f"Could not find a artifact with sample: {sample_id} generated by process: {' ,'.join(process_types)}"
        LOG.error(message)
        raise MissingArtifactError(message=message)

    return specific_result_files



from genologics.entities import Process

from cg_lims.EPPs.arnold.prep import build_step_documents
from cg_lims.models.arnold.base_step import BaseStep

from tests.conftest import (
    server,
)


MICROBIAL_STEP_TYPES = {
    "buffer_exchange",
    "reception_control",
    "normalization_for_sequencing",
    "post_pcr_bead_purification",
    "library_prep",
    "normalization",
}


def test_microbial_prep(lims):
    # GIVEN: A lims with a process: "24-240265" (CG002 - Normalization of microbial samples for sequencing).
    # Where the samples in the process has gone through the whole microbial prep workflow:

    server("microbial_prep")
    process = Process(lims=lims, id="24-240265")

    # WHEN running build_step_documents
    step_documents = build_step_documents(lims=lims, process=process, prep_type="micro")

    # THEN assert BaseStep documents are created and all step types in the microbial workflow are represented
    for document in step_documents:
        assert isinstance(document, BaseStep)
    assert {document.step_type for document in step_documents} == MICROBIAL_STEP_TYPES



"""CLI module to calculate the sample and water volume based on the amount of RNA needed,
as specified in the step `Aliquot Samples for Fragmentation (RNA)` """

import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import InvalidValueError, LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

TOTAL_VOLUME = 25


def calculate_sample_and_water_volumes(artifacts: List[Artifact]):
    """Calculate the sample and water volumes for all samples"""
    missing_udfs = 0
    for artifact in artifacts:
        concentration = artifact.udf.get("Concentration")
        amount_needed = artifact.udf.get("Amount needed (ng)")
        if concentration is None or amount_needed is None:
            LOG.error(
                f"Sample {artifact.samples[0].name} is missing udf 'Concentration' or 'Amount needed'."
            )
            missing_udfs += 1
            continue
        sample_volume = amount_needed / concentration
        artifact.udf["Sample Volume (ul)"] = sample_volume
        artifact.udf["Volume H2O (ul)"] = TOTAL_VOLUME - sample_volume
        artifact.put()

    if missing_udfs:
        raise MissingUDFsError(
            f"Could not apply calculation for {missing_udfs} out of {len(artifacts)} sample(s): "
            f"'Concentration' or 'Amount needed (ng)' is missing!"
        )



@click.command()
@click.pass_context
def calculate_water_volume_rna(context: click.Context):
    """Calculate sample and water volumes based on the amount of RNA needed"""
    LOG.info(f"Running {context.command_path} with params: {context.params}")
    process = context.obj["process"]
    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=False)
        calculate_sample_and_water_volumes(artifacts=artifacts)
        message = "Sample and H2O volumes have been calculated!"
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



"""CLI module for calculating beads volumes"""
import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)


def calculate_water_volume(sample_volume: float, sample_volume_limit: float) -> float:
    """Calculates the H20 volume based on the sample volume"""
    return sample_volume_limit - sample_volume if sample_volume < sample_volume_limit else 0.0


def calculate_volumes(artifacts: List[Artifact], sample_volume_limit: float):
    """Calculates water volume and total volume"""

    missing_udfs = 0
    high_volume_warning = False
    warning_message = ""
    for artifact in artifacts:
        sample_volume: float = artifact.udf.get("Sample Volume (ul)")
        if sample_volume is None:
            missing_udfs += 1
            continue
        h2o_volume = calculate_water_volume(
            sample_volume=sample_volume, sample_volume_limit=sample_volume_limit
        )
        total_volume = h2o_volume + sample_volume
        if total_volume > 100:
            high_volume_warning = True
        artifact.udf["Volume H2O (ul)"] = h2o_volume
        artifact.udf["Total Volume (uL)"] = total_volume
        artifact.put()

    if missing_udfs:
        warning_message += (
            f'Udf "Sample Volume (ul)" missing for {missing_udfs} out of {len(artifacts)} samples '
        )
    if high_volume_warning:
        warning_message += "Total volume higher than 100 ul for some samples!"
    if warning_message:
        raise MissingUDFsError(message=warning_message)


@click.command()
@options.sample_volume_limit()
@click.pass_context
def volume_water(context: click.Context, sample_volume_limit: float):
    """Water volume calculation."""

    LOG.info(f"Running {context.command_path} with params: {context.params}")

    process = context.obj["process"]

    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=False)
        calculate_volumes(artifacts=artifacts, sample_volume_limit=sample_volume_limit)
        message = "Beads volumes have been calculated."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



#!/usr/bin/env python
from typing import Tuple

from genologics.entities import Process, Artifact
from genologics.lims import Lims

import logging
import sys

import click

from cg_lims.exceptions import LimsError

LOG = logging.getLogger(__name__)


def find_reads_to_sequence(process: Process, lims: Lims) -> Tuple[int, int]:
    """Function to copy Missing reads Pool (M) from input artifact if existing,
    otherwise Reads missing (M) from sample."""

    failed_artifacts = 0
    passed_artifacts = 0

    for input, output in process.input_output_maps:
        if output.get("output-type") != "Analyte":
            continue

        input_artifact = Artifact(lims, id=input["limsid"])
        output_artifact = Artifact(lims, id=output["limsid"])
        if input_artifact.udf.get("Missing reads Pool (M)") is not None:
            reads_to_sequence = input_artifact.udf.get("Missing reads Pool (M)")
        elif output_artifact.samples[0].udf.get("Reads missing (M)") is not None:
            reads_to_sequence = output_artifact.samples[0].udf.get("Reads missing (M)")
        else:
            failed_artifacts += 1
            continue

        output_artifact.udf["Reads to sequence (M)"] = str(reads_to_sequence)
        output_artifact.put()
        passed_artifacts += 1
    return passed_artifacts, failed_artifacts


@click.command()
@click.pass_context
def reads_to_sequence(ctx):
    """Script to find Reads to sequence (M) for a sample or a pool"""

    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        find_reads_to_sequence(process=process, lims=lims)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



"""One time script to add Covid meta data to sample"""
import logging
import pathlib
from datetime import date, datetime
from typing import List

import click
import coloredlogs
import pandas as pd
import yaml
from genologics.entities import Sample
from genologics.lims import Lims

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingSampleError

LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]
LOG = logging.getLogger(__name__)
LOG_FORMAT = "%(levelname)s:%(message)s"
ONLY_ADD_IF_DATA_IS_MISSING = [
    "Collection Date",
    "Region",
    "Lab Code",
    "Original Lab",
    "Original Lab Address",
    "Region Code",
]


def get_covid_samples(row: pd.Series, lims: Lims) -> List[Sample]:
    """Tries to find a single sample based on the submitted sample name in the meta data file"""
    sample_name = row["Submitted Sample Name"]
    LOG.info("Trying to find submitted sample %s:", sample_name)
    samples = lims.get_samples(name=sample_name)
    if len(samples) == 0:
        message = f"Sample {sample_name} in meta data file not found in LIMS!"
        LOG.warning(message)
        raise MissingSampleError(message)
    LOG.info("Sample(s) found: %s!", ", ".join([sample.id for sample in samples]))
    return samples


def update_sample_udfs(sample: Sample, udfs: pd.Index, row: pd.Series) -> None:
    """Updates the udfs on a sample with data from the meta data file"""
    for udf in udfs:
        if udf == "Collection Date" and not isinstance(row[udf], date):
            collection_date = row[udf]
            converted_collection_date = datetime.strptime(row[udf], "%Y-%m-%d").date()
            LOG.info(
                "Converting Collection Date from string to date format: %s -> %s",
                collection_date,
                converted_collection_date,
            )
            row[udf] = converted_collection_date
        if udf == "Region Code":
            LOG.info("Converting Region Code from int to string format")
            row[udf] = str(row[udf]).zfill(2)
        if not sample.udf.get(udf):
            LOG.info("Creating udf %s on sample %s: %s", udf, sample.id, row[udf])
        if sample.udf.get(udf) and sample.udf[udf] != row[udf]:
            if udf in ONLY_ADD_IF_DATA_IS_MISSING:
                LOG.warning(
                    "udf %s on sample (%s) differs from meta data (%s), sample will NOT be "
                    "updated!",
                    udf,
                    sample.udf[udf],
                    row[udf],
                )
                continue
            else:
                LOG.info(
                    "Updating udf %s on sample %s: %s -> %s",
                    udf,
                    sample.id,
                    sample.udf[udf],
                    row[udf],
                )
        if sample.udf.get(udf) and sample.udf[udf] == row[udf]:
            LOG.info("Meta data already present on sample %s", sample.id)
        sample.udf[udf] = row[udf]


def set_meta_data_on_samples(meta_data: pd.DataFrame, lims: Lims) -> None:
    """Finds a sample based on the submitted sample name and sets the udfs"""
    successful_updates: int = 0
    failed_updates: int = 0
    udfs: pd.Index = meta_data.columns[1:]
    for _, row in meta_data.iterrows():
        try:
            samples: List[Sample] = get_covid_samples(row, lims)
            for sample in samples:
                LOG.info(
                    "Meta data found for sample %s: Collection Date %s, Region %s, Lab Code %s, "
                    "Original Lab %s, Original Lab Address %s, Region Code %s, Data Analysis %s",
                    sample.id,
                    row["Collection Date"],
                    row["Region"],
                    row["Lab Code"],
                    row["Original Lab"],
                    row["Original Lab Address"],
                    row["Region Code"],
                    row["Data Analysis"],
                )
                update_sample_udfs(sample, udfs, row)
                sample.put()
                successful_updates += 1
        except MissingSampleError:
            failed_updates += 1
    if failed_updates:
        raise LimsError(
            f"Meta data successfully updated on {successful_updates} samples! Number of submitted "
            f"samples not updated: {failed_updates}"
        )


@click.command()
@options.config()
@options.log()
@click.option(
    "--data-file",
    "-f",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="path to file with meta data",
)
@click.option(
    "--log-level",
    type=click.Choice(LEVELS),
    default="INFO",
    help="lowest level to log at",
)
@click.pass_context
def set_sample_meta_data(
    context: click.Context, config: str, log: str, data_file: click.Path, log_level: str
):
    """Script to set Covid meta data on sample udf."""
    with open(config) as file:
        config_data = yaml.load(file, Loader=yaml.FullLoader)
    lims = Lims(
        config_data["BASEURI"], config_data["USERNAME"], config_data["PASSWORD"]
    )
    log_path = pathlib.Path(log)
    log_level = getattr(logging, log_level.upper())
    logging.basicConfig(
        filename=str(log_path.absolute()), filemode="w", level=log_level
    )
    LOG.info(f"Running {context.command_path} with params: {context.params}")
    coloredlogs.install(level=log_level, fmt=LOG_FORMAT)
    meta_data: pd.DataFrame = pd.read_csv(data_file, sep=";")
    try:
        set_meta_data_on_samples(meta_data, lims)
        message = f"All samples updated!"
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        click.echo(e.message)


if __name__ == "__main__":
    set_sample_meta_data()



from typing import Optional

from genologics.lims import Lims
from pydantic.main import BaseModel
from pydantic import Field

from cg_lims.models.arnold.base_step import BaseStep
from cg_lims.objects import BaseAnalyte


class ProcessUDFs(BaseModel):
    """Xp Load to FlowCell (Nova Seq)"""

    workflow_type: Optional[str] = Field(None, alias="Workflow Type")
    paired_end: Optional[str] = Field(None, alias="Paired End")
    index_read_1: Optional[str] = Field(None, alias="Index Read 1")
    index_read_2: Optional[str] = Field(None, alias="Index Read 2")
    read_1_cycles: Optional[str] = Field(None, alias="Read 1 Cycles")
    read_2_cgitycles: Optional[str] = Field(None, alias="Read 2 Cycles")


class ArnoldStep(BaseStep):
    process_udfs: ProcessUDFs

    class Config:
        allow_population_by_field_name = True


def get_xp_load_to_flowcell(lims: Lims, sample_id: str, prep_id: str) -> Optional[ArnoldStep]:
    analyte = BaseAnalyte(
        lims=lims,
        sample_id=sample_id,
        process_type="Xp Load to FlowCell (Nova Seq)",
        optional_step=True,
    )
    if not analyte.process:
        return None
    return ArnoldStep(
        **analyte.base_fields(),
        process_udfs=ProcessUDFs(**analyte.process_udfs()),
        sample_id=sample_id,
        prep_id=prep_id,
        step_type="xp_load_to_flowcell",
        workflow="NovaSeq"
    )



#!/usr/bin/env python
import logging
import sys

import click
from cg_lims.get.processes import get_latest_process
from genologics.entities import Process, Artifact
from genologics.lims import Lims

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts
from cg_lims import options
from typing import List, Dict, Literal, Any

LOG = logging.getLogger(__name__)


def get_copy_tasks(
    process: Process,
) -> Dict[str, Dict[Literal["Source Step", "Source Field"], Any]]:
    copy_tasks = {}

    for udf, value in process.udf.items():
        if "Copy task" not in udf:
            continue
        task_number, copy_type = udf.split("-")
        task_number = task_number.strip()
        copy_type = copy_type.strip()
        if copy_type not in ["Source Step", "Source Field"]:
            message = f'Copy type was: {copy_type}! Must be "Source Step" or "Source Field"'
            LOG.error(message)
            raise MissingUDFsError(message)
        if task_number in copy_tasks:
            copy_tasks[task_number][copy_type] = value
        else:
            copy_tasks[task_number] = {copy_type: value}
    if copy_tasks == {}:
        message = f"Copy task UDF missing for process {process.id}"
        LOG.error(message)
        raise MissingUDFsError(message)
    return copy_tasks


def copy_udfs(input_artifacts: List[Artifact], copy_tasks: dict, lims: Lims, reception_control: List[str]):
    """Loop through all artifacts and copy udfs from the correct steps. QC flags are collected from the
       reception control step if specified and no fitting source processes can be found."""
    failed_udfs = []
    process_types = [task["Source Step"] for task_number, task in copy_tasks.items()]
    process_types = list(set(process_types))
    for artifact in input_artifacts:
        artifact.get()
        source_steps = get_correct_steps_for_the_artifact(
            process_types=process_types, artifact=artifact, lims=lims
        )
        if source_steps == {} and reception_control != ():
            source_steps = get_correct_steps_for_the_artifact(
                process_types=reception_control, artifact=artifact, lims=lims
            )
            current_copy_tasks = {"Copy task 1": {"Source Field": "", "Source Step": reception_control[0]}}
            message = f"No fitting processes found for artifact {artifact.id}. " \
                      f"Retrieving QC flags from {reception_control}."
            LOG.info(message)
        else:
            current_copy_tasks = copy_tasks
        for process_type, source_step in source_steps.items():
            for output in source_step.outputs_per_input(artifact.id):
                qc_flag_update = Artifact(lims, id=output.id).qc_flag
                if not qc_flag_update or qc_flag_update == "UNKNOWN":
                    continue
                if artifact.qc_flag != "FAILED":
                    artifact.qc_flag = qc_flag_update
                for task_number, task in current_copy_tasks.items():
                    if task["Source Step"] != process_type:
                        continue
                    udf = task["Source Field"]
                    udf_value = output.udf.get(udf)
                    if udf_value is None:
                        failed_udfs.append(artifact.name)
                        continue
                    try:
                        artifact.udf[udf] = float(udf_value)
                    except:
                        artifact.udf[udf] = str(udf_value)

        artifact.put()


def get_correct_steps_for_the_artifact(
    process_types: List[str], artifact: Artifact, lims: Lims
) -> Dict[str, Process]:
    """Get the latest processes of the specified process types."""

    source_steps = {}
    processes = lims.get_processes(inputartifactlimsid=artifact.id, type=process_types)
    for process in processes:
        process_type = process.type.name
        if process_type in source_steps:
            source_steps[process_type] = get_latest_process([source_steps[process_type], process])
        else:
            source_steps[process_type] = process
    return source_steps


@click.command()
@options.process_types(help="Use this flag to specify Reception Control process type.")
@click.pass_context
def aggregate_qc_and_copy_fields(ctx, process_types: List[str]):
    """Script to copy artifact udf to sample udf"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]
    try:
        copy_tasks = get_copy_tasks(process=process)
        artifacts = get_artifacts(process=process, input=True)
        copy_udfs(input_artifacts=artifacts, copy_tasks=copy_tasks, lims=lims, reception_control=process_types)
        click.echo("Udfs have been set on all samples.")
    except LimsError as e:
        sys.exit(e.message)



"""
CLI module to calculate the resuspension buffer volume (RB Volume) based on the amount of DNA
needed, as specified in the step `Aliquot Samples for Covaris`

For an explanation of the RB volume calculation, see AM-documents 1057, 1464 and 1717
"""

import logging
import sys
from typing import List

import click
from genologics.entities import Artifact

from cg_lims.exceptions import InvalidValueError, LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

AMOUNT_NEEDED_LUCIGEN = 200
AMOUNT_NEEDED_TRUSEQ = 1100
TOTAL_VOLUME_LUCIGEN = 25
TOTAL_VOLUME_TRUSEQ = 55


def calculate_rb_volume(artifacts: List[Artifact]):
    """Calculate the RB volumes for all samples"""
    missing_udfs = 0
    for artifact in artifacts:
        concentration = artifact.udf.get("Concentration")
        amount_needed = artifact.udf.get("Amount needed (ng)")
        if concentration is None:
            LOG.error(
                f"Sample {artifact.id} is missing udf 'Concentration'."
            )
            missing_udfs += 1
            continue
        if amount_needed not in [AMOUNT_NEEDED_LUCIGEN, AMOUNT_NEEDED_TRUSEQ]:
            raise InvalidValueError(
                f"'Amount needed (ng)' missing or incorrect for one or more samples. Value can "
                f"only be {', '.join(map(str, [AMOUNT_NEEDED_LUCIGEN, AMOUNT_NEEDED_TRUSEQ]))}."
            )
        sample_volume = amount_needed / concentration
        if amount_needed == AMOUNT_NEEDED_LUCIGEN:
            artifact.udf["RB Volume (ul)"] = TOTAL_VOLUME_LUCIGEN - sample_volume
        if amount_needed == AMOUNT_NEEDED_TRUSEQ:
            artifact.udf["RB Volume (ul)"] = TOTAL_VOLUME_TRUSEQ - sample_volume
        artifact.udf["Sample Volume (ul)"] = sample_volume
        artifact.put()

    if missing_udfs:
        raise MissingUDFsError(
            f"Could not apply calculation for {missing_udfs} out of {len(artifacts)} sample(s): "
            f"'Concentration' is missing!"
        )


@click.command()
@click.pass_context
def calculate_resuspension_buffer_volume(context: click.Context):
    """Calculate RB volumes based on the amount of DNA needed"""
    LOG.info(f"Running {context.command_path} with params: {context.params}")
    process = context.obj["process"]
    try:
        artifacts: List[Artifact] = get_artifacts(process=process, input=False)
        calculate_rb_volume(artifacts=artifacts)
        message = "RB volumes have been calculated!"
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)



import logging
import sys
from typing import List, Dict
import click
from datetime import date
from genologics.entities import Artifact, Process
from genologics.lims import Lims
from cg_lims import options
from cg_lims.exceptions import LimsError
from .models import PlacementMapHeader, WellInfo, PlateInfo
from .hmtl_templates import (
    PLACEMENT_MAP_HEADER,
    PLATE_HEADER_SECTION,
    TABLE_HEADERS,
    TABLE_ROWS,
    VISUAL_PLACEMENT_MAP_HEADER,
    VISUAL_PLACEMENT_MAP_WELL,
)

LOG = logging.getLogger(__name__)


def get_placement_map_data(
    lims: Lims, process: Process, udfs: List[str], original_well: str
) -> dict:
    """collecting the data for the placement map."""

    placement_map = {}
    for input, output in process.input_output_maps:
        if output.get("output-generation-type") == "PerAllInputs":
            continue
        input_artifact = Artifact(lims, id=input["limsid"])
        output_artifact = Artifact(lims, id=output["limsid"])
        destination_container = output_artifact.location[0]
        if destination_container:
            if destination_container not in placement_map:
                placement_map[destination_container] = {}
            destination_well = output_artifact.location[1]
            placement_map[destination_container][destination_well] = make_source_dest_info(
                input_artifact, output_artifact, original_well, udfs
            )
    return placement_map


def make_source_dest_info(
    source_artifact: Artifact, dest_artifact: Artifact, original_well: str, udfs: List[str]
) -> WellInfo:
    """Collecting info about a well."""

    sample = source_artifact.samples[0]
    if len(source_artifact.samples) > 1:
        sample_name = source_artifact.name
        sample_id = source_artifact.id
        sample_type = "Pool"
    else:
        sample_name = sample.name
        sample_id = sample.id
        sample_type = "Sample"

    if original_well:
        container_source = "Original Container"
        container_name = sample.udf.get("Original Container", "")
        well_source = "Original Well"
        well = sample.udf.get("Original Well", "")
    else:
        container_source = "Source Container"
        container_name = source_artifact.location[0].name
        well_source = "Source Well"
        well = source_artifact.location[1]
    return WellInfo(
        project_name=sample.project.name,
        sample_name=sample_name,
        sample_id=sample_id,
        sample_type=sample_type,
        container_source=container_source,
        container_name=container_name,
        well_source=well_source,
        well=well,
        exta_udf_info=more_sample_info(source_artifact, udfs),
        dest_well=dest_artifact.location[1],
    )


def more_sample_info(artifact: Artifact, udfs: List[str]) -> str:
    """Adding more udf info about sample in the well."""

    html = []
    for udf in udfs:
        value = artifact.udf.get(udf)
        if value is not None:
            if isinstance(value, float):
                value = round(value, 2)
            html.append(f"{udf} : {value}")
    return "".join(html)


def add_wells(container_info: Dict[str, WellInfo]):
    """Building the plate wells for the visual placement map."""

    html = []
    coulmns = range(1, 14)
    rows = ["A", "B", "C", "D", "E", "F", "G", "H"]
    for rowname in rows:
        html.append(f"""<tr style="height: 12%;"><td class="bold-column row-name">{rowname}</td>""")
        for column in coulmns:
            well_location = f"{rowname}:{column}"
            if well_location in container_info:
                # If there is an artifact in the well:
                well_info = container_info[well_location]
                html.append(VISUAL_PLACEMENT_MAP_WELL.format(**well_info.dict()))
            else:
                # For wells that are empty:
                html.append('<td class="well" style="">&nbsp;</td>')
        html.append("</tr>")
    return "".join(html)


def make_html(placement_map: dict, process: Process, original_well: str) -> str:
    """Creating the html."""

    html = ["<html>"]
    header_info = PlacementMapHeader(process_type=process.type.name, date=date.today().isoformat())
    html.append(PLACEMENT_MAP_HEADER.format(**header_info.dict()))
    container_source = "Original" if original_well else "Source"
    for container, container_info in placement_map.items():
        plate_info = PlateInfo(
            container_name=container.name,
            container_type=container.type.name,
            container_id=container.id,
        )
        html.append("""<table class="group-contents"><br><br><thead>""")
        html.append(PLATE_HEADER_SECTION.format(**plate_info.dict()))
        html.append(TABLE_HEADERS.format(container_source=container_source))
        html.append("</thead>")
        html.append("<tbody>")
        for dest_well, well_data in container_info.items():
            html.append(TABLE_ROWS.format(**well_data.dict()))
        html.append("</tbody></table><br><br>")

        ## VISUAL Platemap

        html.append(VISUAL_PLACEMENT_MAP_HEADER)
        html.append("<tbody>")
        html.append(add_wells(container_info))
        html.append("</tbody></table>")
    html.append("</html>")
    html = "".join(html)
    return html


@click.command()
@options.file_placeholder(help="File placeholder name.")
@options.sample_udfs(help="Sample UDFs to show in the placement map.")
@options.original_well()
@click.pass_context
def placement_map(ctx, file: str, sample_udfs: List[str], original_well: str):
    """Create a 96 well placement map."""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    lims = ctx.obj["lims"]

    try:
        placement_map_data = get_placement_map_data(lims, process, sample_udfs, original_well)
        html = make_html(placement_map_data, process, original_well)
        with open(f"{file}.html", "w") as file:
            file.write(html)
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)



from typing import List, Tuple, Iterator

from genologics.entities import Artifact, Process

import logging

from cg_lims.exceptions import MissingUDFsError

LOG = logging.getLogger(__name__)


def copy_artifact_to_artifact(
    destination_artifact: Artifact,
    source_artifact: Artifact,
    artifact_udfs: List[Tuple[str, str]],
    qc_flag: bool = False,
    keep_failed_flags: bool = False,
) -> None:
    """Copying artifact udfs from source artifact to destination artifact.
    Will also copy qc_flag if set to True
    Logging missing uds. Raising MissingUDFsError if any udf is missing"""

    missing_udfs = 0
    artifacts_to_put = False
    for source_udf, destination_udf in artifact_udfs:
        if source_artifact.udf.get(source_udf) is None:
            message = f"Artifact udf {source_udf} missing on artifact {source_artifact.id}"
            LOG.error(message)
            missing_udfs += 1
            continue
        destination_artifact.udf[destination_udf] = source_artifact.udf[source_udf]
        artifacts_to_put = True

    if qc_flag:
        if keep_failed_flags and destination_artifact.qc_flag == "FAILED":
            message = f"QC for destination artifact {destination_artifact.id} is failed, " \
                      f"flag not copied over from source artifact {source_artifact.id}"
            LOG.error(message)
        else:
            destination_artifact.qc_flag = source_artifact.qc_flag
            artifacts_to_put = True

    if artifacts_to_put:
        destination_artifact.put()
    if missing_udfs:
        raise MissingUDFsError(f"Some UDFs missing on the source artifact: {source_artifact.id}")


def copy_udf_process_to_artifact(
    artifact: Artifact, process: Process, artifact_udf: str, process_udf: str
) -> None:
    """Copying process udf to artifact udf.
    Logging and raising missing uds error"""

    if not process.udf.get(process_udf):
        message = f"process: {process.id} missing udf {process_udf}"
        LOG.error(message)
        raise MissingUDFsError(message=message)
    try:
        artifact.udf[artifact_udf] = process.udf[process_udf]
        artifact.put()
    except:
        message = f"{artifact_udf} doesn't seem to be a valid artifact udf."
        LOG.error(message)
        raise MissingUDFsError(message=message)



from genologics.entities import Process, Processtype

from cg_lims.EPPs.arnold.sequencing import build_step_documents
from cg_lims.models.arnold.base_step import BaseStep

from tests.conftest import (
    server,
)


NOVA_SEQ_STANDARD_STEP_TYPES = {
    "bcl_conversion_and_demultiplexing",
    "define_run_format",
    "standard_make_pool_and_denature",
    "standard_prepare_for_sequencing",
}


def test_nova_seq_standard(lims):
    # GIVEN: A lims with a process: "24-308692" (Bcl Conversion & Demultiplexing (Nova Seq)).
    # Where the samples in the process has gone through the whole nova seq standard protocol:

    server("novaseq_standard")
    process = Process(lims=lims, id="24-308986")

    # WHEN running build_step_documents for the novaseq sequencing_method
    step_documents = build_step_documents(lims=lims, process=process, sequencing_method="novaseq")

    # THEN assert BaseStep documents are created and all required step types in the covid prep workflow are represented
    for document in step_documents:
        assert isinstance(document, BaseStep)
    assert {document.step_type for document in step_documents} == NOVA_SEQ_STANDARD_STEP_TYPES



#!/usr/bin/env python

import click

# commands
from cg_lims.EPPs.udf.check.check_artifact_udfs import check_artifact_udfs


@click.group(invoke_without_command=True)
@click.pass_context
def check(ctx):
    """Main entry point of check commands"""
    pass


check.add_command(check_artifact_udfs)



import logging
from typing import Optional
from pydantic import Field, BaseModel, validator
from datetime import date, datetime

LOG = logging.getLogger(__name__)


class BaseStep(BaseModel):
    """Artifact related fields that are not udfs."""

    prep_id: str
    step_type: str
    sample_id: str
    workflow: str
    lims_step_name: Optional[str]
    id: Optional[str] = Field(alias="_id")
    step_id: Optional[str] = Field(alias="_id")
    well_position: Optional[str]
    artifact_name: Optional[str]
    container_name: Optional[str]
    container_id: Optional[str]
    container_type: Optional[str]
    index_name: Optional[str]
    nr_samples_in_pool: Optional[int]
    date_run: Optional[date]

    @validator("id", always=True)
    def set_id(cls, v, values):
        return f"{values['prep_id']}_{values['step_type']}"

    @validator("step_id", always=True)
    def set_step_id(cls, v, values):
        return f"{values['prep_id']}_{values['step_type']}"

    @validator("date_run", always=True)
    def get_date_run(cls, v, values) -> Optional[str]:
        if isinstance(v, date):
            return datetime(v.year, v.month, v.day).__str__()
        return None



#!/usr/bin/env python
import logging
import sys
from pathlib import Path
from typing import List

import click
from genologics.lims import Artifact

from cg_lims import options
from cg_lims.EPPs.files.hamilton.models import BarcodeFileRow
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.files.manage_csv_files import build_csv, sort_csv, sort_csv_plate_and_tube
from cg_lims.get.artifacts import get_artifacts

LOG = logging.getLogger(__name__)

HEADERS = [
    "Source Labware",
    "Barcode Source Container",
    "Source Well",
    "Sample Volume",
    "Destination Labware",
    "Barcode Destination Container",
    "Destination Well",
    "Buffer Volume",
]


def get_file_data_and_write(
    destination_artifacts: List[Artifact], file: str, volume_udf: str, buffer_udf: str, pool: bool
):
    """Making a hamilton normalization file with sample and buffer volumes, source and destination barcodes and wells."""

    failed_samples = []
    missing_source_barcode = []
    missing_destination_barcode = []
    file_rows = []
    for destination_artifact in destination_artifacts:
        if destination_artifact.udf.get("Output Container Barcode") is None:
            missing_destination_barcode.append(destination_artifact.id)
        source_artifacts = destination_artifact.input_artifact_list()
        buffer = True
        for source_artifact in source_artifacts:
            try:
                row_data = BarcodeFileRow(
                    source_artifact=source_artifact,
                    destination_artifact=destination_artifact,
                    pool=pool,
                    buffer=buffer,
                    sample_volume=volume_udf,
                    buffer_volume=buffer_udf,
                )
                if pool:
                    buffer = False
            except:
                failed_samples.append(source_artifact.id)
                continue

            row_data_dict = row_data.dict(by_alias=True)

            file_rows.append([row_data_dict[header] for header in HEADERS])

            if source_artifact.udf.get("Output Container Barcode") is None:
                missing_source_barcode.append(source_artifact.id)

    build_csv(file=Path(file), rows=file_rows, headers=HEADERS)
    sort_csv_plate_and_tube(file=Path(file),
                            plate_columns=["Barcode Source Container", "Source Well"],
                            tube_columns=["Destination Well"],
                            plate_well_columns=["Source Well"],
                            tube_well_columns=["Destination Well"])

    if failed_samples:
        raise MissingUDFsError(
            f"All samples were not added to the file. Udfs missing for samples: {', '.join(failed_samples)}"
        )
    if missing_source_barcode or missing_destination_barcode:
        raise MissingUDFsError(
            f"Barcodes missing for some artifacts. "
            f"Missing source barcode: {', '.join(missing_source_barcode)}. "
            f"Missing destination barcode: {', '.join(missing_destination_barcode)}."
        )


@click.command()
@options.file_placeholder(help="Hamilton Noramlization File")
@options.buffer_udf()
@options.volume_udf()
@options.pooling_step()
@click.pass_context
def barcode_file(
    ctx: click.Context, file: str, volume_udf: str, buffer_udf: str, pooling_step: bool
):
    """Script to make a hamilton normalization file"""

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    artifacts = get_artifacts(process=process, input=False)
    try:
        get_file_data_and_write(
            pool=pooling_step,
            destination_artifacts=artifacts,
            file=f"{file}-hamilton-normalization.txt",
            volume_udf=volume_udf,
            buffer_udf=buffer_udf,
        )
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)



#!/usr/bin/env python

import logging
import sys

import click
from typing import List
from genologics.entities import Artifact

from cg_lims import options
from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.files.manage_csv_files import make_plate_file
from cg_lims.get.artifacts import get_artifacts
from cg_lims.get.fields import get_artifact_well
from cg_lims.get.samples import get_one_sample_from_artifact

LOG = logging.getLogger(__name__)


def get_file_data_and_write(artifacts: List[Artifact], file: str) -> None:
    """Getting row data for hamilton file from artifacts"""

    file_rows = {}
    for artifact in artifacts:
        well = get_artifact_well(artifact=artifact)
        if well == "11":
            message = "No well found for sample(s). Script does not work without a sample plate."
            LOG.error(message)
            raise MissingUDFsError(message=message)

        lims_id = get_one_sample_from_artifact(artifact=artifact).id

        barcode = artifact.udf.get("Input Container Barcode", "-")
        sample_volume = artifact.udf.get("Sample Volume (ul)", "-")
        beads_volume = artifact.udf.get("Volume Beads (ul)", "-")
        elution_volume = artifact.udf.get("Volume Elution (ul)", "-")

        file_rows[well] = [
            barcode,
            lims_id,
            well,
            sample_volume,
            beads_volume,
            elution_volume,
        ]

    headers = [
        "Input Plate Barcode",
        "LIMS ID",
        "Well ID",
        "Volume Sample",
        "Volume Beads",
        "Volume Elution buffer",
    ]

    make_plate_file(file, file_rows, headers)


def resolve_file_extension(extension: str) -> str:
    """Appends a period symbol to extension string if none is present. Otherwise, returns given extension string"""
    if extension:
        if extension.startswith("."):
            return extension
        return f".{extension}"
    return ""


@click.command()
@options.file_placeholder(help="Hamilton file.")
@options.file_extension()
@click.pass_context
def buffer_exchange_twist_file(ctx: click.Context, file: str, extension: str):
    """Script to make a csv file for hamilton for step Buffer Exchange TWIST. See AM doc #2125"""
    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")
    process = ctx.obj["process"]
    artifacts = get_artifacts(process=process, input=False)

    file_name = (
        f"{file}_{artifacts[0].container.name}_{process.type.name.replace(' ', '_')}"
        f"{resolve_file_extension(extension)}"
    )

    try:
        get_file_data_and_write(artifacts, file_name)
        click.echo("The file was successfully generated.")
    except LimsError as e:
        sys.exit(e.message)



import logging
import sys
from typing import List, Optional

import click
from genologics.entities import Artifact

from cg_lims.exceptions import LimsError, MissingUDFsError
from cg_lims.get.artifacts import get_artifacts
from cg_lims import options
from cg_lims.set.qc import set_qc_fail

LOG = logging.getLogger(__name__)


def calculate_concentrations(
    artifacts: List[Artifact],
    size_udf: str,
    molar_concentration_udf: str,
    concentration_udf: str,
    upper_threshold: Optional[float] = None,
    lower_threshold: Optional[float] = None,
) -> None:
    """
    Formula to calculate molar concentration (nM) from weight concentration (ng/ul) and fragment size (bp) is:

    molar_concentration = concentration * 1e6 / (average_molecular_weight * strands * fragment_size)

    where:
    - fragment_size is the number of base pairs of the DNA fragment (bp)
    - concentration is the concentration of dna (ng/ul)
    - average_molecular_weight = 328.3 is the average molecular weight of nucleotides, in g/mol.
    - strands = (one/two) for single or double stranded DNA
    - molar_concentration (nM)
    """
    average_molecular_weight = 328.3
    strands = 2
    missing_udfs_count = 0

    for artifact in artifacts:
        size = artifact.udf.get(size_udf)
        concentration = artifact.udf.get(concentration_udf)
        if None in [size, concentration]:
            missing_udfs_count += 1
            continue
        factor = 1e6 / (average_molecular_weight * strands * float(size))
        concentration_nm = concentration * factor
        artifact.udf[molar_concentration_udf] = concentration_nm
        sample = artifact.samples[0]
        if (
            upper_threshold
            and sample.name[0:3] == "NTC"
            and sample.udf["Sequencing Analysis"][0:2] == "MW"
        ):
            set_qc_fail(
                artifact=artifact, value=concentration_nm, threshold=upper_threshold, criteria=">="
            )
        elif lower_threshold:
            set_qc_fail(
                artifact=artifact, value=concentration_nm, threshold=lower_threshold, criteria="<="
            )
        artifact.put()
    if missing_udfs_count:
        passed_artifacts = len(artifacts) - missing_udfs_count
        raise MissingUDFsError(
            f"Udf missing for {missing_udfs_count} sample(s). Concentration calculated for {passed_artifacts} samples."
        )


@click.command()
@options.input()
@options.measurement()
@options.size_udf()
@options.concantration_nm_udf()
@options.concantration_udf()
@options.lower_threshold()
@options.upper_threshold()
@click.pass_context
def molar_concentration(
    ctx,
    input: bool,
    measurement: Optional[bool],
    size_udf: str,
    conc_nm_udf: str,
    conc_udf: str,
    lower_threshold: Optional[float],
    upper_threshold: Optional[float],
) -> None:
    """Script to calculate molar concentration given the weight concentration and fragment size. """

    LOG.info(f"Running {ctx.command_path} with params: {ctx.params}")

    process = ctx.obj["process"]

    try:
        artifacts = get_artifacts(process=process, input=input, measurement=measurement)
        calculate_concentrations(
            artifacts=artifacts,
            size_udf=size_udf,
            molar_concentration_udf=conc_nm_udf,
            concentration_udf=conc_udf,
            lower_threshold=lower_threshold,
            upper_threshold=upper_threshold,
        )
        message = "Concentration (nM) have been calculated for all samples."
        LOG.info(message)
        click.echo(message)
    except LimsError as e:
        LOG.error(e.message)
        sys.exit(e.message)
